
class Ram(object):
    def __init__(self, instr, input_registers):
        self.instr = instr # list of instructions
        self.current = 0 # current instruction
        self.input_registers = input_registers # I registers
        self.output_registers = [] # O registers
        self.work_registers = [] # R registers

    def read_register(self, type_register, index, ref_origin=None, ref_target=None):
        '''
        Read from a register
        '''
        if type_register == "i": # input
            return self.input_registers[index]
        elif type_register == "r": # work
            value = self.work_registers[index]
            if value is None:
                raise ValueError("register empty")
            else:
                return self.work_registers[index]
        elif type_register == "o": # output
            raise TypeError("output registers are write only")
        elif type_register == "value":
            return index
        elif type_register == "@": # reference
            target_index = self.read_register(ref_origin, index)
            return self.read_register(ref_target, target_index)

    def write_register(self, value, type_register, index, ref_origin=None, ref_target=None):
        '''
        Write to a register
        '''
        if type_register == "i": # input
            raise TypeError("input registers are read only")
        elif type_register == "r": # work
            if len(self.work_registers) <= index:
                self.work_registers.extend([None] * (index + 1 - len(self.work_registers))) # extend with uninitialized values
            self.work_registers[index] = value
        elif type_register == "o": #output
            if len(self.output_registers) <= index:
                self.output_registers.extend([None] * (index + 1 - len(self.output_registers))) # extend with uninitialized values
            self.output_registers[index] = value
        elif type_register == "value":
            raise TypeError("cannot write on value")
        elif type_register == "@": # reference
            target_index = self.read_register(ref_origin, index)
            self.write_register(ref_target, target_index, value)

    def op(self, type_op, r1, r2, r3):
        self.current += 1
        if type_op == 'ADD':
            value_1 = self.read_register(*r1)
            value_2 = self.read_register(*r2)
            value_3 = value_1 + value_2
            self.write_register(value_3, *r3)
        elif type_op == 'SUB':
            value_1 = self.read_register(*r1)
            value_2 = self.read_register(*r2)
            value_3 = value_1 - value_2
            self.write_register(value_3, *r3)
        elif type_op == 'DIV':
            value_1 = self.read_register(*r1)
            value_2 = self.read_register(*r2)
            value_3 = value_1 / value_2
            self.write_register(value_3, *r3)
        elif type_op == 'MULT':
            value_1 = self.read_register(*r1)
            value_2 = self.read_register(*r2)
            value_3 = value_1 * value_2
            self.write_register(value_3, *r3)

    def op_ctrl(self, type_op, z, r1=None, r2=None):
        if type_op == 'JUMP':
            if isinstance(z, int):
                self.current += z
            else:
               raise ValueError("wrong operand type (should be integer)")
        if type_op == 'JE':
            value_1 = self.read_register(*r1)
            value_2 = self.read_register(*r2)
            if value_1 == value_2:
                self.op_ctrl('JUMP', z)
            else:
                self.current += 1
        if type_op == 'JL':
            value_1 = self.read_register(*r1)
            value_2 = self.read_register(*r2)
            if value_1 > value_2:
                self.op_ctrl('JUMP', z)
            else:
                self.current += 1

    def execute(self):
        while self.current < len(self.instr):
            print(self.current)
            print(*self.instr[self.current]['args'])
            self.instr[self.current]['op'](self, *self.instr[self.current]['args'])



### example
# input_registers = [10, 5, 1]
#
# instructions = [
#     {'op': Ram.op, 'args': ('ADD', ('i', 0), ('value', 1), ('o', 2))},
#     {'op': Ram.op, 'args': ('ADD', ('value', 0), ('@', 2, 'i', 'i'), ('o', 1))},
# ]

# input_registers = [10, 5, 1, 2, 4]
# instructions = [ {"op": Ram.op, "args": ('ADD', ('i', 0), ('value', 1), ('r', 0))},
# {"op": Ram.op, "args": ('SUB', ('i', 1), ('value', 3), ('r', 1))},
# {"op": Ram.op, "args": ('MULT', ('r', 1), ('r', 0), ('o', 1))},
# ]


# ram = Ram(instructions, input_registers)

# ram.execute()

# print("Result:", ram.output_registers)
